const std = @import("std");
const Allocator = std.mem.Allocator;

const sdl = @import("sdl");
const Window = @import("./window.zig");
const makeSdlError = Window.makeSdlError;

const Canvas = @This();

allocator: Allocator,
surface: *sdl.SDL_Surface,
pixels: []Pixel,
width: usize,
height: usize,

pub fn init(allocator: Allocator, width: usize, height: usize) !*Canvas {
    const surface = sdl.SDL_CreateRGBSurfaceWithFormat(
        0,
        @intCast(width),
        @intCast(height),
        0,
        sdl.SDL_PIXELFORMAT_ABGR8888,
    ) orelse
        return makeSdlError();

    const pixels_ptr: [*]Pixel = @alignCast(@ptrCast(surface.pixels));
    const pixels_len = width * height;
    const pixels = pixels_ptr[0..pixels_len];

    const result = try allocator.create(Canvas);
    result.* = .{
        .allocator = allocator,
        .surface = surface,
        .pixels = pixels,
        .width = width,
        .height = height,
    };
    return result;
}

pub fn deinit(self: *Canvas) void {
    sdl.SDL_FreeSurface(self.surface);
    self.allocator.destroy(self);
}

pub fn getUnsafe(self: *Canvas, x: usize, y: usize) *Pixel {
    const index = x + y * self.width;
    return &self.pixels[index];
}

pub fn get(self: *Canvas, x: usize, y: usize) *Pixel {
    std.debug.assert(x < self.width);
    std.debug.assert(y < self.height);
    return self.getUnsafe(x, y);
}

pub fn put(self: *Canvas, x: usize, y: usize, value: Pixel) void {
    self.get(x, y).* = value;
}

pub fn putRect(self: *Canvas, x: usize, y: usize, w: usize, h: usize, value: Pixel) void {
    std.debug.assert(x + w <= self.width);
    std.debug.assert(y + h <= self.height);
    const start_index = x + y * self.width;
    for (0..h) |yo| {
        const row_index = start_index + yo * self.width;
        @memset(self.pixels[row_index..][0..w], value);
    }
}

pub fn putSprite(self: *Canvas, x: usize, y: usize, sprite: *const Sprite) void {
    std.debug.assert(x + 16 <= self.width);
    std.debug.assert(y + 16 <= self.height);
    for (0..16) |yo|
        for (0..16) |xo| {
            const value = sprite[yo][xo];
            if (value.a == 0x00) continue;

            const dest = self.getUnsafe(x + xo, y + yo);
            if (value.a == 0xFF)
                dest.* = value
            else
                @panic("Not supported");
        };
}

pub fn clear(self: *Canvas, value: Pixel) void {
    @memset(self.pixels, value);
}

pub const Pixel = packed struct {
    r: u8,
    g: u8,
    b: u8,
    a: u8,

    pub const transparent = Pixel{ .r = 0x00, .g = 0x00, .b = 0x00, .a = 0x00 };
    pub const black = Pixel{ .r = 0x00, .g = 0x00, .b = 0x00, .a = 0xFF };
    pub const white = Pixel{ .r = 0xFF, .g = 0xFF, .b = 0xFF, .a = 0xFF };
};

pub const Sprite = [16][16]Pixel;
