const std = @import("std");
const Allocator = std.mem.Allocator;

const sdl = @import("sdl");
const Window = @import("./window.zig");
const makeSdlError = Window.makeSdlError;

const Canvas = @import("./canvas.zig");
const Pixel = Canvas.Pixel;
const Sprite = Canvas.Sprite;

const Grid = @import("./grid.zig");

pub fn main() !void {
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    const allocator = gpa.allocator();
    var prng = std.Random.DefaultPrng.init(0);
    const random = prng.random();

    var grid = Grid{};
    const key_layer = grid.layerAs(0, u8);

    var sprite_lookup: [256]Sprite = undefined;
    try loadSprites(&sprite_lookup);

    const size = Grid.size * Grid.size;

    const scale = 3;
    const window = try Window.init(allocator, size * scale, size * scale, key_layer);
    defer window.deinit();

    const canvas = try Canvas.init(allocator, size, size);
    defer canvas.deinit();

    while (window.running) {
        window.pollEvents();

        grid.get(Grid.Pos.init(-5, -4, 0)).as(u8).* +%= 1;
        grid.get(Grid.Pos.init(-4, -4, 0)).as(u8).* = random.int(u8);

        render(&grid, canvas, &sprite_lookup);
        try window.updateSurface(canvas);
    }
}

fn render(grid: *Grid, canvas: *Canvas, sprite_lookup: *[256]Sprite) void {
    canvas.clear(Pixel.black);

    for (0..Grid.size) |vx|
        for (0..Grid.size) |vy|
            for (0..Grid.size) |vz| {
                const x = @as(isize, @intCast(vx)) - Grid.size / 2;
                const y = @as(isize, @intCast(vy)) - Grid.size / 2;
                const z = @as(isize, @intCast(vz)) - Grid.size / 2;

                const pos = Grid.Pos.initTruncate(x, y, z);
                const value = grid.get(pos).as(u8).*;

                const sprite = &sprite_lookup[value];
                canvas.putSprite(vx * Grid.size, vy * Grid.size, sprite);
            };
}

fn loadSprites(sprite_lookup: *[256]Sprite) !void {
    const png_surface = sdl.IMG_Load("./gfx/sprite_map.png") orelse
        return makeSdlError();
    defer sdl.SDL_FreeSurface(png_surface);

    std.debug.assert(png_surface.w == 256);
    std.debug.assert(png_surface.h == 256);

    const pixels_surface = sdl.SDL_ConvertSurfaceFormat(
        png_surface,
        sdl.SDL_PIXELFORMAT_ABGR8888,
        0,
    ) orelse
        return makeSdlError();
    defer sdl.SDL_FreeSurface(pixels_surface);

    const pixels_ptr: [*]Pixel = @alignCast(@ptrCast(pixels_surface.pixels));
    const pixels_len: usize = @intCast(pixels_surface.w * pixels_surface.h);
    const pixels = pixels_ptr[0..pixels_len];

    for (sprite_lookup, 0..) |*sprite, i| {
        const x = i % 16;
        const y = i / 16;
        for (0..16) |yo| {
            const dst_index = x * 16 + (y * 16 + yo) * 256;
            @memcpy(&sprite[yo], pixels[dst_index..][0..16]);
        }
    }
}
